@@ -15,7 +15,7 @@ from tests import unittest
 import mock
 
 from botocore.stub import Stubber
-from botocore.exceptions import ParamValidationError, StubResponseError
+from botocore.exceptions import ParamValidationError, StubResponseError, UnStubbedResponseError
 from botocore.model import ServiceModel
 from botocore import hooks
 
@@ -68,6 +68,23 @@ class TestStubber(unittest.TestCase):
         self.event_emitter.unregister.assert_any_call(
             'before-call.*.*', mock.ANY, unique_id=mock.ANY)
 
+    def test_context_manager(self):
+        self.event_emitter = mock.Mock()
+        self.client.meta.events = self.event_emitter
+
+        with self.stubber:
+            # Ensure events are registered in context
+            self.event_emitter.register_first.assert_called_with(
+                'before-parameter-build.*.*', mock.ANY, unique_id=mock.ANY)
+            self.event_emitter.register.assert_called_with(
+                'before-call.*.*', mock.ANY, unique_id=mock.ANY)
+
+        # Ensure events are no longer registered once we leave the context
+        self.event_emitter.unregister.assert_any_call(
+            'before-parameter-build.*.*', mock.ANY, unique_id=mock.ANY)
+        self.event_emitter.unregister.assert_any_call(
+            'before-call.*.*', mock.ANY, unique_id=mock.ANY)
+
     def test_add_response(self):
         response = {'foo': 'bar'}
         self.stubber.add_response('foo', response)
@@ -142,9 +159,42 @@ class TestStubber(unittest.TestCase):
         self.assertEqual(response[1]['Error']['Message'], service_message)
         self.assertEqual(response[1]['Error']['Code'], error_code)
 
+    def test_get_client_error_with_extra_error_meta(self):
+        error_code = "foo"
+        error_message = "bar"
+        error_meta = {
+            "Endpoint": "https://foo.bar.baz",
+        }
+        self.stubber.add_client_error(
+            'foo', error_code, error_message,
+            http_status_code=301,
+            service_error_meta=error_meta)
+        with self.stubber:
+            response = self.emit_get_response_event()
+        error = response[1]['Error']
+        self.assertIn('Endpoint', error)
+        self.assertEqual(error['Endpoint'], "https://foo.bar.baz")
+
+    def test_get_client_error_with_extra_response_meta(self):
+        error_code = "foo"
+        error_message = "bar"
+        stub_response_meta = {
+            "RequestId": "79104EXAMPLEB723",
+        }
+        self.stubber.add_client_error(
+            'foo', error_code, error_message,
+            http_status_code=301,
+            response_meta=stub_response_meta)
+        with self.stubber:
+            response = self.emit_get_response_event()
+        actual_response_meta = response[1]['ResponseMetadata']
+        self.assertIn('RequestId', actual_response_meta)
+        self.assertEqual(actual_response_meta['RequestId'], "79104EXAMPLEB723")
+
+
     def test_get_response_errors_with_no_stubs(self):
         self.stubber.activate()
-        with self.assertRaises(StubResponseError):
+        with self.assertRaises(UnStubbedResponseError):
             self.emit_get_response_event()
 
     def test_assert_no_responses_remaining(self):
