@@ -11,6 +11,7 @@
 # ANY KIND, either express or implied. See the License for the specific
 # language governing permissions and limitations under the License.
 from tests import unittest
+from tests import RawResponse
 from dateutil.tz import tzutc, tzoffset
 import datetime
 import copy
@@ -21,11 +22,11 @@ from botocore import xform_name
 from botocore.compat import OrderedDict, json
 from botocore.compat import six
 from botocore.awsrequest import AWSRequest
+from botocore.awsrequest import AWSResponse
 from botocore.exceptions import InvalidExpressionError, ConfigNotFound
-from botocore.exceptions import ClientError
+from botocore.exceptions import ClientError, ConnectionClosedError
 from botocore.exceptions import InvalidDNSNameError, MetadataRetrievalError
 from botocore.model import ServiceModel
-from botocore.vendored import requests
 from botocore.utils import is_json_value_header
 from botocore.utils import remove_dot_segments
 from botocore.utils import normalize_url_path
@@ -45,6 +46,7 @@ from botocore.utils import fix_s3_host
 from botocore.utils import switch_to_virtual_host_style
 from botocore.utils import instance_cache
 from botocore.utils import merge_dicts
+from botocore.utils import lowercase_dict
 from botocore.utils import get_service_module_name
 from botocore.utils import percent_encode_sequence
 from botocore.utils import percent_encode
@@ -176,6 +178,12 @@ class TestTransformName(unittest.TestCase):
         transformed = xform_name('IPV6', '_')
         self.assertEqual(transformed, 'ipv6')
 
+    def test_s3_partial_rename(self):
+        transformed = xform_name('s3Resources', '-')
+        self.assertEqual(transformed, 's3-resources')
+        transformed = xform_name('s3Resources', '_')
+        self.assertEqual(transformed, 's3_resources')
+
 
 class TestValidateJMESPathForSet(unittest.TestCase):
     def setUp(self):
@@ -984,6 +992,33 @@ class TestMergeDicts(unittest.TestCase):
             dict1, {'Foo': ['foo_value']})
 
 
+class TestLowercaseDict(unittest.TestCase):
+    def test_lowercase_dict_empty(self):
+        original = {}
+        copy = lowercase_dict(original)
+        self.assertEqual(original, copy)
+
+    def test_lowercase_dict_original_keys_lower(self):
+        original = {
+            'lower_key1': 1,
+            'lower_key2': 2,
+        }
+        copy = lowercase_dict(original)
+        self.assertEqual(original, copy)
+
+    def test_lowercase_dict_original_keys_mixed(self):
+        original = {
+            'SOME_KEY': 'value',
+            'AnOTher_OnE': 'anothervalue',
+        }
+        copy = lowercase_dict(original)
+        expected = {
+            'some_key': 'value',
+            'another_one': 'anothervalue',
+        }
+        self.assertEqual(expected, copy)
+
+
 class TestGetServiceModuleName(unittest.TestCase):
     def setUp(self):
         self.service_description = {
@@ -1434,6 +1469,26 @@ class TestS3RegionRedirector(unittest.TestCase):
             request_dict, response, self.operation)
         self.assertIsNone(redirect_response)
 
+    def test_does_not_redirect_400_head_bucket_no_region_header(self):
+        # We should not redirect a 400 Head* if the region header is not
+        # present as this will lead to infinitely calling HeadBucket.
+        request_dict = {'url': 'https://us-west-2.amazonaws.com/foo',
+                        'context': {'signing': {'bucket': 'foo'}}}
+        response = (None, {
+            'Error': {'Code': '400', 'Message': 'Bad Request'},
+            'ResponseMetadata': {
+                'HTTPHeaders': {}
+            }
+        })
+
+        self.operation.name = 'HeadBucket'
+        redirect_response = self.redirector.redirect_from_error(
+            request_dict, response, self.operation)
+        head_bucket_calls = self.client.head_bucket.call_count
+        self.assertIsNone(redirect_response)
+        # We should not have made an additional head bucket call
+        self.assertEqual(head_bucket_calls, 0)
+
     def test_does_not_redirect_if_None_response(self):
         request_dict = {'url': 'https://us-west-2.amazonaws.com/foo',
                         'context': {'signing': {'bucket': 'foo'}}}
@@ -1521,7 +1576,7 @@ class TestContainerMetadataFetcher(unittest.TestCase):
     def fake_response(self, status_code, body):
         response = mock.Mock()
         response.status_code = status_code
-        response.text = body
+        response.content = body
         return response
 
     def set_http_responses_to(self, *responses):
@@ -1535,9 +1590,15 @@ class TestContainerMetadataFetcher(unittest.TestCase):
                 http_response = response
             else:
                 http_response = self.fake_response(
-                    status_code=200, body=json.dumps(response))
+                    status_code=200, body=json.dumps(response).encode('utf-8'))
             http_responses.append(http_response)
-        self.http.get.side_effect = http_responses
+        self.http.send.side_effect = http_responses
+
+    def assert_request(self, method, url, headers):
+        request = self.http.send.call_args[0][0]
+        self.assertEqual(request.method, method)
+        self.assertEqual(request.url, url)
+        self.assertEqual(request.headers, headers)
 
     def assert_can_retrieve_metadata_from(self, full_uri):
         response_body = {'foo': 'bar'}
@@ -1545,10 +1606,7 @@ class TestContainerMetadataFetcher(unittest.TestCase):
         fetcher = self.create_fetcher()
         response = fetcher.retrieve_full_uri(full_uri)
         self.assertEqual(response, response_body)
-        self.http.get.assert_called_with(
-            full_uri, headers={'Accept': 'application/json'},
-            timeout=fetcher.TIMEOUT_SECONDS,
-        )
+        self.assert_request('GET', full_uri, {'Accept': 'application/json'})
 
     def assert_host_is_not_allowed(self, full_uri):
         response_body = {'foo': 'bar'}
@@ -1556,7 +1614,7 @@ class TestContainerMetadataFetcher(unittest.TestCase):
         fetcher = self.create_fetcher()
         with self.assertRaisesRegexp(ValueError, 'Unsupported host'):
             fetcher.retrieve_full_uri(full_uri)
-        self.assertFalse(self.http.get.called)
+        self.assertFalse(self.http.send.called)
 
     def test_can_specify_extra_headers_are_merged(self):
         headers = {
@@ -1569,10 +1627,7 @@ class TestContainerMetadataFetcher(unittest.TestCase):
         fetcher = self.create_fetcher()
         response = fetcher.retrieve_full_uri(
             'http://localhost', headers)
-        self.http.get.assert_called_with(
-            'http://localhost', headers=headers,
-            timeout=fetcher.TIMEOUT_SECONDS,
-        )
+        self.assert_request('GET', 'http://localhost', headers)
 
     def test_can_retrieve_uri(self):
         json_body =  {
@@ -1588,11 +1643,8 @@ class TestContainerMetadataFetcher(unittest.TestCase):
 
         self.assertEqual(response, json_body)
         # Ensure we made calls to the right endpoint.
-        self.http.get.assert_called_with(
-            'http://169.254.170.2/foo?id=1',
-            headers={'Accept': 'application/json'},
-            timeout=fetcher.TIMEOUT_SECONDS,
-        )
+        headers = {'Accept': 'application/json'}
+        self.assert_request('GET', 'http://169.254.170.2/foo?id=1', headers)
 
     def test_can_retry_requests(self):
         success_response = {
@@ -1604,7 +1656,7 @@ class TestContainerMetadataFetcher(unittest.TestCase):
         self.set_http_responses_to(
             # First response is a connection error, should
             # be retried.
-            requests.ConnectionError(),
+            ConnectionClosedError(endpoint_url=''),
             # Second response is the successful JSON response
             # with credentials.
             success_response,
@@ -1616,44 +1668,44 @@ class TestContainerMetadataFetcher(unittest.TestCase):
     def test_propagates_credential_error_on_http_errors(self):
         self.set_http_responses_to(
             # In this scenario, we never get a successful response.
-            requests.ConnectionError(),
-            requests.ConnectionError(),
-            requests.ConnectionError(),
-            requests.ConnectionError(),
-            requests.ConnectionError(),
+            ConnectionClosedError(endpoint_url=''),
+            ConnectionClosedError(endpoint_url=''),
+            ConnectionClosedError(endpoint_url=''),
+            ConnectionClosedError(endpoint_url=''),
+            ConnectionClosedError(endpoint_url=''),
         )
         # As a result, we expect an appropriate error to be raised.
         fetcher = self.create_fetcher()
         with self.assertRaises(MetadataRetrievalError):
             fetcher.retrieve_uri('/foo?id=1')
-        self.assertEqual(self.http.get.call_count, fetcher.RETRY_ATTEMPTS)
+        self.assertEqual(self.http.send.call_count, fetcher.RETRY_ATTEMPTS)
 
     def test_error_raised_on_non_200_response(self):
         self.set_http_responses_to(
-            self.fake_response(status_code=404, body='Error not found'),
-            self.fake_response(status_code=404, body='Error not found'),
-            self.fake_response(status_code=404, body='Error not found'),
+            self.fake_response(status_code=404, body=b'Error not found'),
+            self.fake_response(status_code=404, body=b'Error not found'),
+            self.fake_response(status_code=404, body=b'Error not found'),
         )
         fetcher = self.create_fetcher()
         with self.assertRaises(MetadataRetrievalError):
             fetcher.retrieve_uri('/foo?id=1')
         # Should have tried up to RETRY_ATTEMPTS.
-        self.assertEqual(self.http.get.call_count, fetcher.RETRY_ATTEMPTS)
+        self.assertEqual(self.http.send.call_count, fetcher.RETRY_ATTEMPTS)
 
     def test_error_raised_on_no_json_response(self):
         # If the service returns a sucess response but with a body that
         # does not contain JSON, we should still retry up to RETRY_ATTEMPTS,
         # but after exhausting retries we propagate the exception.
         self.set_http_responses_to(
-            self.fake_response(status_code=200, body='Not JSON'),
-            self.fake_response(status_code=200, body='Not JSON'),
-            self.fake_response(status_code=200, body='Not JSON'),
+            self.fake_response(status_code=200, body=b'Not JSON'),
+            self.fake_response(status_code=200, body=b'Not JSON'),
+            self.fake_response(status_code=200, body=b'Not JSON'),
         )
         fetcher = self.create_fetcher()
         with self.assertRaises(MetadataRetrievalError):
             fetcher.retrieve_uri('/foo?id=1')
         # Should have tried up to RETRY_ATTEMPTS.
-        self.assertEqual(self.http.get.call_count, fetcher.RETRY_ATTEMPTS)
+        self.assertEqual(self.http.send.call_count, fetcher.RETRY_ATTEMPTS)
 
     def test_can_retrieve_full_uri_with_fixed_ip(self):
         self.assert_can_retrieve_metadata_from(
@@ -1700,55 +1752,231 @@ class TestUnsigned(unittest.TestCase):
 
 class TestInstanceMetadataFetcher(unittest.TestCase):
     def setUp(self):
-        self._requests_patch = mock.patch('botocore.utils.requests')
-        self._requests = self._requests_patch.start()
-
-    def tearDown(self):
-        self._requests_patch.stop()
-
-    def test_disabled_by_environment(self):
-        env = {'AWS_EC2_METADATA_DISABLED': 'true'}
-        fetcher = InstanceMetadataFetcher(env=env)
-        result = fetcher.retrieve_iam_role_credentials()
-        self.assertEqual(result, {})
-        self._requests.assert_not_called()
-
-    def test_disabled_by_environment_mixed_case(self):
-        env = {'AWS_EC2_METADATA_DISABLED': 'tRuE'}
-        fetcher = InstanceMetadataFetcher(env=env)
-        result = fetcher.retrieve_iam_role_credentials()
-        self.assertEqual(result, {})
-        self._requests.get.assert_not_called()
-
-    def test_disabling_env_var_not_true(self):
-        url = 'https://example.com/'
-        env = {'AWS_EC2_METADATA_DISABLED': 'false'}
-        creds = {
+        urllib3_session_send = 'botocore.httpsession.URLLib3Session.send'
+        self._urllib3_patch = mock.patch(urllib3_session_send)
+        self._send = self._urllib3_patch.start()
+        self._imds_responses = []
+        self._send.side_effect = self.get_imds_response
+        self._role_name = 'role-name'
+        self._creds = {
             'AccessKeyId': 'spam',
             'SecretAccessKey': 'eggs',
             'Token': 'spam-token',
             'Expiration': 'something',
         }
 
-        profiles_response = mock.Mock()
-        profiles_response.status_code = 200
-        profiles_response.content = b'role-name'
+    def tearDown(self):
+        self._urllib3_patch.stop()
 
-        creds_response = mock.Mock()
-        creds_response.status_code = 200
-        creds_response.content = json.dumps(creds).encode('utf-8')
+    def add_imds_response(self, body, status_code=200):
+        response = botocore.awsrequest.AWSResponse(
+            url='http://169.254.169.254/',
+            status_code=status_code,
+            headers={},
+            raw=RawResponse(body)
+        )
+        self._imds_responses.append(response)
 
-        self._requests.get.side_effect = [profiles_response, creds_response]
+    def add_get_role_name_imds_response(self, role_name=None):
+        if role_name is None:
+            role_name = self._role_name
+        self.add_imds_response(body=role_name.encode('utf-8'))
+
+    def add_get_credentials_imds_response(self, creds=None):
+        if creds is None:
+            creds = self._creds
+        self.add_imds_response(body=json.dumps(creds).encode('utf-8'))
+
+    def add_imds_connection_error(self, exception):
+        self._imds_responses.append(exception)
+
+    def get_imds_response(self, request):
+        response = self._imds_responses.pop(0)
+        if isinstance(response, Exception):
+            raise response
+        return response
+
+    def test_disabled_by_environment(self):
+        env = {'AWS_EC2_METADATA_DISABLED': 'true'}
+        fetcher = InstanceMetadataFetcher(env=env)
+        result = fetcher.retrieve_iam_role_credentials()
+        self.assertEqual(result, {})
+        self._send.assert_not_called()
+
+    def test_disabled_by_environment_mixed_case(self):
+        env = {'AWS_EC2_METADATA_DISABLED': 'tRuE'}
+        fetcher = InstanceMetadataFetcher(env=env)
+        result = fetcher.retrieve_iam_role_credentials()
+        self.assertEqual(result, {})
+        self._send.assert_not_called()
+
+    def test_disabling_env_var_not_true(self):
+        url = 'https://example.com/'
+        env = {'AWS_EC2_METADATA_DISABLED': 'false'}
+
+        self.add_get_role_name_imds_response()
+        self.add_get_credentials_imds_response()
 
         fetcher = InstanceMetadataFetcher(url=url, env=env)
         result = fetcher.retrieve_iam_role_credentials()
 
         expected_result = {
-            'access_key': 'spam',
-            'secret_key': 'eggs',
-            'token': 'spam-token',
-            'expiry_time': 'something',
-            'role_name': 'role-name'
+            'access_key': self._creds['AccessKeyId'],
+            'secret_key': self._creds['SecretAccessKey'],
+            'token': self._creds['Token'],
+            'expiry_time': self._creds['Expiration'],
+            'role_name': self._role_name
         }
         self.assertEqual(result, expected_result)
 
+    def test_includes_user_agent_header(self):
+        user_agent = 'my-user-agent'
+        self.add_get_role_name_imds_response()
+        self.add_get_credentials_imds_response()
+
+        InstanceMetadataFetcher(
+            user_agent=user_agent).retrieve_iam_role_credentials()
+
+        headers = self._send.call_args[0][0].headers
+        self.assertEqual(headers['User-Agent'], user_agent)
+
+    def test_non_200_response_for_role_name_is_retried(self):
+        # Response for role name that have a non 200 status code should
+        # be retried.
+        self.add_imds_response(
+            status_code=429, body=b'{"message": "Slow down"}')
+        self.add_get_role_name_imds_response()
+        self.add_get_credentials_imds_response()
+        result = InstanceMetadataFetcher(
+            num_attempts=2).retrieve_iam_role_credentials()
+        expected_result = {
+            'access_key': self._creds['AccessKeyId'],
+            'secret_key': self._creds['SecretAccessKey'],
+            'token': self._creds['Token'],
+            'expiry_time': self._creds['Expiration'],
+            'role_name': self._role_name
+        }
+        self.assertEqual(result, expected_result)
+
+    def test_http_connection_error_for_role_name_is_retried(self):
+        # Connection related errors should be retried
+        self.add_imds_connection_error(ConnectionClosedError(endpoint_url=''))
+        self.add_get_role_name_imds_response()
+        self.add_get_credentials_imds_response()
+        result = InstanceMetadataFetcher(
+            num_attempts=2).retrieve_iam_role_credentials()
+        expected_result = {
+            'access_key': self._creds['AccessKeyId'],
+            'secret_key': self._creds['SecretAccessKey'],
+            'token': self._creds['Token'],
+            'expiry_time': self._creds['Expiration'],
+            'role_name': self._role_name
+        }
+        self.assertEqual(result, expected_result)
+
+    def test_empty_response_for_role_name_is_retried(self):
+        # Response for role name that have a non 200 status code should
+        # be retried.
+        self.add_imds_response(body=b'')
+        self.add_get_role_name_imds_response()
+        self.add_get_credentials_imds_response()
+        result = InstanceMetadataFetcher(
+            num_attempts=2).retrieve_iam_role_credentials()
+        expected_result = {
+            'access_key': self._creds['AccessKeyId'],
+            'secret_key': self._creds['SecretAccessKey'],
+            'token': self._creds['Token'],
+            'expiry_time': self._creds['Expiration'],
+            'role_name': self._role_name
+        }
+        self.assertEqual(result, expected_result)
+
+    def test_non_200_response_is_retried(self):
+        self.add_get_role_name_imds_response()
+        # Response for creds that has a 200 status code but has an empty
+        # body should be retried.
+        self.add_imds_response(
+            status_code=429, body=b'{"message": "Slow down"}')
+        self.add_get_credentials_imds_response()
+        result = InstanceMetadataFetcher(
+            num_attempts=2).retrieve_iam_role_credentials()
+        expected_result = {
+            'access_key': self._creds['AccessKeyId'],
+            'secret_key': self._creds['SecretAccessKey'],
+            'token': self._creds['Token'],
+            'expiry_time': self._creds['Expiration'],
+            'role_name': self._role_name
+        }
+        self.assertEqual(result, expected_result)
+
+    def test_http_connection_errors_is_retried(self):
+        self.add_get_role_name_imds_response()
+        # Connection related errors should be retried
+        self.add_imds_connection_error(ConnectionClosedError(endpoint_url=''))
+        self.add_get_credentials_imds_response()
+        result = InstanceMetadataFetcher(
+            num_attempts=2).retrieve_iam_role_credentials()
+        expected_result = {
+            'access_key': self._creds['AccessKeyId'],
+            'secret_key': self._creds['SecretAccessKey'],
+            'token': self._creds['Token'],
+            'expiry_time': self._creds['Expiration'],
+            'role_name': self._role_name
+        }
+        self.assertEqual(result, expected_result)
+
+    def test_empty_response_is_retried(self):
+        self.add_get_role_name_imds_response()
+        # Response for creds that has a 200 status code but is empty.
+        # This should be retried.
+        self.add_imds_response(body=b'')
+        self.add_get_credentials_imds_response()
+        result = InstanceMetadataFetcher(
+            num_attempts=2).retrieve_iam_role_credentials()
+        expected_result = {
+            'access_key': self._creds['AccessKeyId'],
+            'secret_key': self._creds['SecretAccessKey'],
+            'token': self._creds['Token'],
+            'expiry_time': self._creds['Expiration'],
+            'role_name': self._role_name
+        }
+        self.assertEqual(result, expected_result)
+
+    def test_invalid_json_is_retried(self):
+        self.add_get_role_name_imds_response()
+        # Response for creds that has a 200 status code but is invalid JSON.
+        # This should be retried.
+        self.add_imds_response(body=b'{"AccessKey":')
+        self.add_get_credentials_imds_response()
+        result = InstanceMetadataFetcher(
+            num_attempts=2).retrieve_iam_role_credentials()
+        expected_result = {
+            'access_key': self._creds['AccessKeyId'],
+            'secret_key': self._creds['SecretAccessKey'],
+            'token': self._creds['Token'],
+            'expiry_time': self._creds['Expiration'],
+            'role_name': self._role_name
+        }
+        self.assertEqual(result, expected_result)
+
+    def test_exhaust_retries_on_role_name_request(self):
+        self.add_imds_response(status_code=400, body=b'')
+        result = InstanceMetadataFetcher(
+            num_attempts=1).retrieve_iam_role_credentials()
+        self.assertEqual(result, {})
+
+    def test_exhaust_retries_on_credentials_request(self):
+        self.add_get_role_name_imds_response()
+        self.add_imds_response(status_code=400, body=b'')
+        result = InstanceMetadataFetcher(
+            num_attempts=1).retrieve_iam_role_credentials()
+        self.assertEqual(result, {})
+
+    def test_missing_fields_in_credentials_response(self):
+        self.add_get_role_name_imds_response()
+        # Response for creds that has a 200 status code and a JSON body
+        # representing an error. We do not necessarily want to retry this.
+        self.add_imds_response(
+            body=b'{"Code":"AssumeRoleUnauthorizedAccess","Message":"error"}')
+        result = InstanceMetadataFetcher().retrieve_iam_role_credentials()
+        self.assertEqual(result, {})
